class Variable:
    """Variable definition in state class."""

    def __init__(self, validate=None):
        self.validate = validate or _default

    def __repr__(self):
        if self.validate is _default:
            return "Variable()"
        else:
            return f"Variable({self.validate!r})"


def _default(value):
    return value



"""Service objects designed with OOP in mind."""
from _stories.actor import Actor
from _stories.argument import Argument
from _stories.initiate import initiate
from _stories.state import State
from _stories.story import Story
from _stories.stubs import I
from _stories.union import Union
from _stories.variable import Variable


__all__ = ("Story", "I", "initiate", "State", "Union", "Argument", "Variable", "Actor")



from _stories.argument import Argument
from _stories.state import State
from _stories.variable import Variable


def Union(*states):
    """Create a union of states.

    State union combine variables and arguments defined in each state.

    Validators for repeated variables and arguments would be combined as well.

    """
    class_name = "Union(" + ", ".join(state.__name__ for state in states) + ")"
    scope = {
        name: lookup[is_argument](_Merge(validators))
        for name, (is_argument, validators) in _compose(states)
    }
    return type(class_name, (State,), scope)


class _Merge:
    def __new__(cls, validators):
        if len(validators) > 1:
            return object.__new__(cls)
        else:
            return validators[0]

    def __init__(self, validators):
        self.validators = validators

    def __call__(self, value):
        for validator in self.validators:
            validated = validator(value)
        return validated


def _compose(states):
    new = {}
    for state in states:
        init = state.__init__
        arguments = init.arguments
        for name, validator in init.validators.items():
            _set(new, name, name in arguments, validator)
    return new.items()


def _set(new, name, is_argument, validator):
    value = new.setdefault(name, [False, []])
    value[0] = value[0] or is_argument
    if isinstance(validator, _Merge):
        value[1].extend(validator.validators)
    else:
        value[1].append(validator)


lookup = {True: Argument, False: Variable}



from _stories.execute import _Executor
from _stories.step import _Step


class _StoryType(type):
    def __prepare__(class_name, bases):
        return {"I": _Step()}

    def __new__(cls, class_name, bases, namespace):
        steps = namespace.pop("I").steps
        if not bases:
            return type.__new__(cls, class_name, bases, namespace)
        namespace["__call__"] = _Executor(steps)
        return type.__new__(cls, class_name, bases, namespace)


class Story(metaclass=_StoryType):
    """Business process specification.

    Use sentences from business domain to express its steps.

    """



from dataclasses import make_dataclass
from inspect import isclass

from _stories.exceptions import StoryError
from _stories.story import Story


def initiate(cls):
    """Create story with all steps required in constructor argument."""
    _check_bases(cls)
    _check_steps(cls)
    _check_init(cls)
    return make_dataclass(
        cls.__name__, cls.__call__.steps, namespace={"__call__": cls.__call__}
    )


def _check_bases(cls):
    if not (isclass(cls) and issubclass(cls, Story)):
        raise StoryError("@initiate can decorate Story subclasses only")


def _check_steps(cls):
    for attrname in cls.__call__.steps:
        if attrname in cls.__dict__:
            raise StoryError("Story decorated by @initiate can not have step methods")


def _check_init(cls):
    if "__init__" in cls.__dict__:
        raise StoryError(
            "Story decorated by @initiate can not have constructor defined"
        )



from asyncio import iscoroutinefunction
from functools import partial
from types import MethodType

from _stories.execute import coroutine
from _stories.execute import function


class _Executor:
    def __init__(self, steps):
        self.steps = steps

    def __get__(self, instance, klass):
        if instance is None:
            return self
        first_step = getattr(instance, self.steps[0])
        if isinstance(first_step, MethodType):
            step = first_step.__func__
        else:
            step = first_step.__call__.func
        if iscoroutinefunction(step):
            func = coroutine._execute
        else:
            func = function._execute
        return partial(func, self.steps, instance)



from _stories.argument import Argument
from _stories.exceptions import StateError
from _stories.variable import Variable


def _setter(state, name, value):
    object.__setattr__(state, name, value)


def _initiator(state, **arguments):
    for argument, value in arguments.items():
        _setter(state, argument, value)


class _ValidateSetter:
    def __init__(self, validators):
        self.validators = validators

    def __get__(self, state, state_class):
        return _BoundValidateSetter(self.validators, state)


class _ValidateInitiator:
    def __init__(self, validators, arguments):
        self.validators = validators
        self.arguments = arguments

    def __get__(self, state, state_class):
        return _BoundValidateInitiator(self.validators, self.arguments, state)


class _BoundValidateSetter:
    def __init__(self, validators, state):
        self.validators = validators
        self.state = state

    def __call__(self, name, value):
        if name not in self.validators:
            message = unknown_variable_template.format(variable=name, state=self.state)
            raise StateError(message)
        validator = self.validators[name]
        validated = validator(value)
        _setter(self.state, name, validated)


class _BoundValidateInitiator:
    def __init__(self, validators, arguments, state):
        self.validators = validators
        self.arguments = arguments
        self.state = state

    def __call__(self, **arguments):
        template = unknown_argument_template
        for argument, value in arguments.items():
            if argument not in self.arguments:
                message = template.format(argument=argument, state=self.state)
                raise StateError(message)
            validator = self.validators[argument]
            validated = validator(value)
            _setter(self.state, argument, validated)


def _representation(state):
    return state.__class__.__name__


def _class_representation(state_class):
    validators = state_class.__init__.validators
    variables = "\n".join(
        f"    {name} = Variable({validators[name]!r})" for name in sorted(validators)
    )
    return f"{state_class.__name__}:\n{variables}"


def _new_validate_state(namespace):
    validators = {
        k: v.validate
        for k, v in namespace.items()
        if isinstance(v, (Variable, Argument))
    }
    arguments = {k for k, v in namespace.items() if isinstance(v, Argument)}
    scope = {
        "__setattr__": _ValidateSetter(validators),
        "__init__": _ValidateInitiator(validators, arguments),
        "__repr__": _representation,
    }
    return scope


class _StateType(type):
    def __new__(cls, class_name, bases, namespace):
        if bases:
            scope = _new_validate_state(namespace)
        else:
            scope = {"__init__": _initiator}
        return type.__new__(cls, class_name, bases, scope)

    def __repr__(cls):
        return _class_representation(cls)


class State(metaclass=_StateType):
    """Business process state."""


unknown_variable_template = """
Unknown variable assignment: {variable}

{state!r}
""".strip()


unknown_argument_template = """
Unknown argument passed: {argument}

{state!r}
""".strip()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src', 'stories': 'src/stories'}

packages = \
['_stories', '_stories.execute', 'stories']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'stories',
    'version': '5.1.0',
    'description': 'Service objects designed with OOP in mind.',
    'long_description': '# Stories [![build](https://img.shields.io/github/workflow/status/proofit404/stories/release?style=flat-square)](https://github.com/proofit404/stories/actions/workflows/release.yml?query=branch%3Arelease) [![pypi](https://img.shields.io/pypi/v/stories?style=flat-square)](https://pypi.org/project/stories)\n\nService objects designed with OOP in mind.\n\n**[Documentation](https://proofit404.github.io/stories) |\n[Source Code](https://github.com/proofit404/stories) |\n[Task Tracker](https://github.com/proofit404/stories/issues)**\n\nA paragraph of text explaining the goal of the library…\n\n## Pros\n\n- A feature\n- B feature\n- etc\n\n## Example\n\nA line of text explaining snippet below…\n\n```pycon\n\n>>> from dataclasses import dataclass\n>>> from typing import Callable\n>>> from stories import Story, I, State\n>>> from app.repositories import load_order, load_customer, create_payment\n\n>>> @dataclass\n... class Purchase(Story):\n...     I.find_order\n...     I.find_customer\n...     I.check_balance\n...     I.persist_payment\n...\n...     def find_order(self, state):\n...         state.order = self.load_order(state.order_id)\n...\n...     def find_customer(self, state):\n...         state.customer = self.load_customer(state.customer_id)\n...\n...     def check_balance(self, state):\n...         if not state.order.affordable_for(state.customer):\n...             raise Exception\n...\n...     def persist_payment(self, state):\n...         state.payment = self.create_payment(\n...             order_id=state.order_id, customer_id=state.customer_id\n...         )\n...\n...     load_order: Callable\n...     load_customer: Callable\n...     create_payment: Callable\n\n>>> purchase = Purchase(\n...     load_order=load_order,\n...     load_customer=load_customer,\n...     create_payment=create_payment,\n... )\n\n>>> state = State(order_id=1, customer_id=1)\n\n>>> purchase(state)\n\n>>> state  # doctest: +SKIP\n\n>>> state.payment.was_received()\nFalse\n\n```\n\n## Questions\n\nIf you have any questions, feel free to create an issue in our\n[Task Tracker](https://github.com/proofit404/stories/issues). We have the\n[question label](https://github.com/proofit404/stories/issues?q=is%3Aopen+is%3Aissue+label%3Aquestion)\nexactly for this purpose.\n\n## Enterprise support\n\nIf you have an issue with any version of the library, you can apply for a paid\nenterprise support contract. This will guarantee you that no breaking changes\nwill happen to you. No matter how old version you\'re using at the moment. All\nnecessary features and bug fixes will be backported in a way that serves your\nneeds.\n\nPlease contact [proofit404@gmail.com](mailto:proofit404@gmail.com) if you\'re\ninterested in it.\n\n## License\n\n`stories` library is offered under the two clause BSD license.\n\n<p align="center">&mdash; ⭐ &mdash;</p>\n',
    'author': 'Josiah Kaviani',
    'author_email': 'proofit404@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/stories',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
}


setup(**setup_kwargs)
